/*
 * jSite2 - Core.java -
 * Copyright © 2008 David Roden
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package net.pterodactylus.jsite.core;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.List;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.pterodactylus.util.logging.Logging;

/**
 * The core of jSite.
 *
 * @author David ‘Bombe’ Roden &lt;bombe@freenetproject.org&gt;
 */
public class CoreImpl implements Core, NodeListener, RequestListener, InsertListener {

	/** The logger. */
	private static final Logger logger = Logging.getLogger(CoreImpl.class.getName());

	/** The core listener support. */
	private final CoreListenerSupport coreListenerSupport = new CoreListenerSupport();

	/** Thread pool. */
	private Executor threadPool = Executors.newCachedThreadPool();

	/** The project manager. */
	private ProjectManager projectManager;

	/** The node manager. */
	private NodeManager nodeManager;

	/** The request manager. */
	/* TODO - remove */
	@SuppressWarnings("unused")
	private RequestManager requestManager;

	/** The insert manager. */
	private InsertManager insertManager;

	//
	// LISTENER MANAGEMENT
	//

	/**
	 * {@inheritDoc}
	 */
	public void addCoreListener(CoreListener coreListener) {
		logger.log(Level.FINEST, "adding core listener: " + coreListener);
		coreListenerSupport.addListener(coreListener);
	}

	/**
	 * {@inheritDoc}
	 */
	public void removeCoreListener(CoreListener coreListener) {
		logger.log(Level.FINEST, "removing core listener: " + coreListener);
		coreListenerSupport.removeListener(coreListener);
	}

	//
	// ACCESSORS
	//

	/**
	 * {@inheritDoc}
	 */
	public Executor getThreadPool() {
		return threadPool;
	}

	/**
	 * Returns the project manager.
	 *
	 * @return The project manager
	 */
	public ProjectManager getProjectManager() {
		return projectManager;
	}

	/**
	 * Sets the project manager to use.
	 *
	 * @param projectManager
	 *            The project manager to use
	 */
	public void setProjectManager(ProjectManager projectManager) {
		this.projectManager = projectManager;
	}

	/**
	 * Returns the node manager.
	 *
	 * @return The node manager
	 */
	public NodeManager getNodeManager() {
		return nodeManager;
	}

	/**
	 * Sets the node manager to use.
	 *
	 * @param nodeManager
	 *            The node manager to use
	 */
	public void setNodeManager(NodeManager nodeManager) {
		this.nodeManager = nodeManager;
	}

	/**
	 * Sets the request manager to use.
	 *
	 * @param requestManager
	 *            The request manager to use
	 */
	public void setRequestManager(RequestManager requestManager) {
		this.requestManager = requestManager;
	}

	/**
	 * Sets the insert manager to use.
	 *
	 * @param insertManager
	 *            The insert manager to use
	 */
	public void setInsertManager(InsertManager insertManager) {
		this.insertManager = insertManager;
	}

	/**
	 * {@inheritDoc}
	 */
	public List<Node> getNodes() {
		return nodeManager.getNodes();
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean isNodeConnected(Node node) {
		return nodeManager.hasNode(node);
	}

	/**
	 * {@inheritDoc}
	 */
	public List<Project> getProjects() {
		return projectManager.getProjects();
	}

	//
	// ACTIONS
	//

	/**
	 * {@inheritDoc}
	 */
	public void start() {
		logger.log(Level.INFO, "starting core…");
		try {
			nodeManager.load();
			coreListenerSupport.fireLoadingNodesDone(nodeManager.getDirectory());
		} catch (IOException ioe1) {
			coreListenerSupport.fireLoadingNodesFailed(nodeManager.getDirectory(), ioe1);
		}
		try {
			projectManager.load();
			coreListenerSupport.fireLoadingProjectsDone(projectManager.getDirectory());
		} catch (IOException ioe1) {
			coreListenerSupport.fireLoadingProjectsFailed(projectManager.getDirectory(), ioe1);
		}
		coreListenerSupport.fireCoreLoaded();
	}

	/**
	 * {@inheritDoc}
	 */
	public void stop() {
		logger.log(Level.INFO, "stopping core…");
		try {
			projectManager.save();
			coreListenerSupport.fireSavingProjectsDone(projectManager.getDirectory());
		} catch (IOException ioe1) {
			coreListenerSupport.fireSavingProjectsFailed(projectManager.getDirectory(), ioe1);
		}
		try {
			nodeManager.save();
			coreListenerSupport.fireSavingNodesDone(nodeManager.getDirectory());
		} catch (IOException ioe1) {
			coreListenerSupport.fireSavingNodesFailed(nodeManager.getDirectory(), ioe1);
		}
		coreListenerSupport.fireCoreStopped();
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean addNode(Node node) throws UnknownHostException {
		logger.log(Level.FINEST, "addNode(node=" + node + ")");
		return nodeManager.addNode(node);
	}

	/**
	 * {@inheritDoc}
	 */
	public void removeNode(Node node) {
		logger.log(Level.FINEST, "removeNode(node=" + node + ")");
		nodeManager.removeNode(node);
	}

	/**
	 * {@inheritDoc}
	 */
	public void connectToNode(Node node) {
		logger.log(Level.FINEST, "connectToNode(node=" + node + ")");
		coreListenerSupport.fireNodeConnecting(node);
		nodeManager.connect(node);
	}

	/**
	 * {@inheritDoc}
	 */
	public void disconnectFromNode(Node node) {
		logger.log(Level.FINEST, "disconnectFromNode(node=" + node + ")");
		nodeManager.disconnect(node);
	}

	/**
	 * {@inheritDoc}
	 */
	public void createProject() throws IOException, JSiteException {
		logger.log(Level.FINEST, "createProject()");
		Project newProject = projectManager.createProject();
		coreListenerSupport.fireProjectAdded(newProject);
	}

	/**
	 * {@inheritDoc}
	 */
	public void insertProject(Node node, Project project) {
		logger.log(Level.FINEST, "insertProject(node=" + node + ", project=" + project + ")");
		insertManager.insertProject(project, node);
	}

	/**
	 * {@inheritDoc}
	 */
	public void cloneProject(Project project) {
		logger.log(Level.FINEST, "cloneProject(project=" + project + ")");
		Project projectClone = projectManager.cloneProject(project);
		coreListenerSupport.fireProjectCloned(project, projectClone);
	}

	/**
	 * {@inheritDoc}
	 */
	public void removeProject(Project project) {
		logger.log(Level.FINEST, "removeProject(project=" + project + ")");
		projectManager.removeProject(project);
		coreListenerSupport.fireProjectRemoved(project);
	}

	//
	// PRIVATE METHODS
	//

	/**
	 * Loads the configuration.
	 */
	@SuppressWarnings("unused")
	private void loadConfig() {
		logger.log(Level.FINEST, "loadConfig()");
		/* TODO */
	}

	/**
	 * Saves the configuration.
	 */
	@SuppressWarnings("unused")
	private void saveConfig() {
		logger.log(Level.FINEST, "saveConfig()");
		/* TODO */
	}

	//
	// INTERFACE NodeListener
	//

	/**
	 * {@inheritDoc}
	 */
	public void nodeAdded(Node node) {
		coreListenerSupport.fireNodeAdded(node);
	}

	/**
	 * {@inheritDoc}
	 */
	public void nodeRemoved(Node node) {
		coreListenerSupport.fireNodeRemoved(node);
	}

	/**
	 * {@inheritDoc}
	 */
	public void nodeConnected(Node node) {
		coreListenerSupport.fireNodeConnected(node);
	}

	/**
	 * {@inheritDoc}
	 */
	public void nodeConnectionFailed(Node node, Throwable cause) {
		coreListenerSupport.fireNodeConnectionFailed(node, cause);
	}

	/**
	 * {@inheritDoc}
	 */
	public void nodeDisconnected(Node node, Throwable throwable) {
		coreListenerSupport.fireNodeDisconnected(node, throwable);
	}

	//
	// INTERFACE RequestListener
	//

	/**
	 * {@inheritDoc}
	 */
	public void requestAdded(Request request) {
		/* TODO - find project and process request */
	}

	/**
	 * @see net.pterodactylus.jsite.core.RequestListener#requestProgressed(Request)
	 */
	public void requestProgressed(Request request) {
		/* TODO - find project and process request */
	}

	/**
	 * @see net.pterodactylus.jsite.core.RequestListener#requestRemoved(net.pterodactylus.jsite.core.Request)
	 */
	public void requestRemoved(Request request) {
		/* TODO - find project and process request */
	}

	/**
	 * @see net.pterodactylus.jsite.core.RequestListener#requestGeneratedURI(net.pterodactylus.jsite.core.Request,
	 *      java.lang.String)
	 */
	public void requestGeneratedURI(Request request, String uri) {
		/* TODO - find project and process request */
	}

	/**
	 * @see net.pterodactylus.jsite.core.RequestListener#requestFinished(net.pterodactylus.jsite.core.Request)
	 */
	public void requestFinished(Request request) {
		/* TODO - find project and process request */
	}

	//
	// INTERFACE InsertListener
	//

	/**
	 * {@inheritDoc}
	 */
	public void insertAdded(Insert insert) {
		/* TODO */
	}

	/**
	 * {@inheritDoc}
	 */
	public void insertRemoved(Insert insert) {
		/* TODO */
	}

	/**
	 * {@inheritDoc}
	 */
	public void insertStarted(Insert insert) {
		/* TODO */
	}

	/**
	 * {@inheritDoc}
	 */
	public void insertProgressed(Insert insert) {
		/* TODO */
	}

	/**
	 * {@inheritDoc}
	 */
	public void insertGeneratedURI(Insert insert, String uri) {
		/* TODO */
	}

	/**
	 * {@inheritDoc}
	 */
	public void insertFinished(Insert insert) {
		/* TODO */
	}

}
