/*
 * jSite2 - RequestManager.java -
 * Copyright © 2008 David Roden
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package net.pterodactylus.jsite.core;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.pterodactylus.fcp.highlevel.GetRequestResult;
import net.pterodactylus.fcp.highlevel.HighLevelCallback;
import net.pterodactylus.fcp.highlevel.HighLevelCallbackListener;
import net.pterodactylus.fcp.highlevel.HighLevelClient;
import net.pterodactylus.fcp.highlevel.HighLevelException;
import net.pterodactylus.fcp.highlevel.HighLevelProgress;
import net.pterodactylus.fcp.highlevel.HighLevelProgressListener;
import net.pterodactylus.fcp.highlevel.PutDirRequestResult;
import net.pterodactylus.fcp.highlevel.PutRequestResult;
import net.pterodactylus.fcp.highlevel.RequestListResult;
import net.pterodactylus.fcp.highlevel.RequestResult;
import net.pterodactylus.jsite.core.Request.Type;
import net.pterodactylus.util.logging.Logging;

/**
 * The request manager keeps track of all the request on all connected nodes.
 * The request manager is added to the {@link NodeManager} as a
 * {@link NodeListener} so that it can fire request-removed events in case a
 * node is disconnected.
 *
 * @author David ‘Bombe’ Roden &lt;bombe@freenetproject.org&gt;
 */
public class RequestManager implements NodeListener, HighLevelProgressListener {

	/** Logger. */
	private static final Logger logger = Logging.getLogger(RequestManager.class.getName());

	/** Request listeners. */
	private List<RequestListener> requestListeners = Collections.synchronizedList(new ArrayList<RequestListener>());

	/** The node manager. */
	private NodeManager nodeManager;

	/** Request lists for all nodes. */
	private Map<Node, Map<String, Request>> nodeRequests = Collections.synchronizedMap(new HashMap<Node, Map<String, Request>>());

	//
	// EVENT MANAGEMENT
	//

	/**
	 * Adds a request listener.
	 *
	 * @param requestListener
	 *            The request listener to add
	 */
	public void addRequestListener(RequestListener requestListener) {
		requestListeners.add(requestListener);
	}

	/**
	 * Removes a request listener.
	 *
	 * @param requestListener
	 *            The request listener to remove
	 */
	public void removeRequestListener(RequestListener requestListener) {
		requestListeners.remove(requestListener);
	}

	/**
	 * Notifies all listeners that a request was added.
	 *
	 * @param request
	 *            The request that was added
	 */
	private void fireRequestAdded(Request request) {
		for (RequestListener requestListener : requestListeners) {
			requestListener.requestAdded(request);
		}
	}

	/**
	 * Notifies all listeners that a request progressed.
	 *
	 * @param request
	 *            The request
	 */
	private void fireRequestProgressed(Request request) {
		for (RequestListener requestListener : requestListeners) {
			requestListener.requestProgressed(request);
		}
	}

	/**
	 * Notifies all listeners that a request was removed.
	 *
	 * @param request
	 *            The request that was removed
	 */
	private void fireRequestRemoved(Request request) {
		for (RequestListener requestListener : requestListeners) {
			requestListener.requestRemoved(request);
		}
	}

	/**
	 * Notifies all listeners that a request has finished.
	 *
	 * @see RequestListener#requestFinished(Request)
	 * @param request
	 *            The request that has finished
	 */
	private void fireRequestFinished(Request request) {
		for (RequestListener requestListener : requestListeners) {
			requestListener.requestFinished(request);
		}
	}

	/**
	 * Notifies all listeners that a request has generated a URI.
	 *
	 * @see RequestListener#requestGeneratedURI(Request, String)
	 * @param request
	 *            The request that has generated a URI
	 * @param uri
	 *            The generated URI
	 */
	private void fireRequestGeneratedURI(Request request, String uri) {
		for (RequestListener requestListener : requestListeners) {
			requestListener.requestGeneratedURI(request, uri);
		}
	}

	//
	// ACCESSORS
	//

	/**
	 * Sets the node manager to use.
	 *
	 * @param nodeManager
	 *            The node manager
	 */
	public void setNodeManager(NodeManager nodeManager) {
		this.nodeManager = nodeManager;
	}

	//
	// ACTIONS
	//

	//
	// PRIVATE ACTIONS
	//

	/**
	 * Requests a list of all running requests from a node. This method will
	 * block until the request has been sent!
	 *
	 * @param node
	 *            The node to get all requests for
	 * @throws IOException
	 *             if an I/O error occurs while communicating with the node
	 * @throws JSiteException
	 *             if there is a problem with the node
	 */
	private void getRequests(final Node node) throws IOException, JSiteException {
		HighLevelClient highLevelClient = nodeManager.getHighLevelClient(node);
		if (highLevelClient == null) {
			logger.log(Level.WARNING, "no client for node: " + node);
			return;
		}
		final Map<String, Request> identifierRequests = Collections.synchronizedMap(new HashMap<String, Request>());
		nodeRequests.put(node, identifierRequests);
		HighLevelCallback<RequestListResult> requestListCallback;
		try {
			requestListCallback = highLevelClient.getRequests();
			requestListCallback.addHighLevelCallbackListener(new HighLevelCallbackListener<RequestListResult>() {

				@SuppressWarnings("synthetic-access")
				public void gotResult(HighLevelCallback<RequestListResult> highLevelCallback) {
					RequestListResult requestListResult;
					try {
						requestListResult = highLevelCallback.getResult();
					} catch (InterruptedException e) {
						logger.log(Level.SEVERE, "getResult() blocked and was interrupted");
						return;
					}
					for (RequestResult requestResult : requestListResult) {
						String identifier = requestResult.getIdentifier();
						logger.log(Level.FINER, "got identifier: " + identifier);
						Request request = identifierRequests.get(identifier);
						if (request == null) {
							request = new Request(node, identifier);
						}
						if (request.getType() == null) {
							if (requestResult instanceof GetRequestResult) {
								request.setType(Type.get);
							} else if (requestResult instanceof PutRequestResult) {
								request.setType(Type.put);
							} else if (requestResult instanceof PutDirRequestResult) {
								request.setType(Type.putDir);
							}
						}
						if (identifierRequests.containsKey(identifier)) {
							continue;
						}
						identifierRequests.put(requestResult.getIdentifier(), request);
						fireRequestAdded(request);
					}
				}
			});
		} catch (HighLevelException hle1) {
			throw new BackendException(hle1);
		}
	}

	//
	// INTERFACE NodeListener
	//

	/**
	 * {@inheritDoc}
	 */
	public void nodeAdded(Node node) {
		HighLevelClient highLevelClient = nodeManager.getHighLevelClient(node);
		if (highLevelClient == null) {
			logger.warning("got nodeAdded but no high-level client: " + node);
			return;
		}
		highLevelClient.addHighLevelProgressListener(this);
	}

	/**
	 * {@inheritDoc}
	 */
	public void nodeRemoved(Node node) {
		HighLevelClient highLevelClient = nodeManager.getHighLevelClient(node);
		if (highLevelClient == null) {
			logger.warning("got nodeRemoved but no high-level client: " + node);
			return;
		}
		highLevelClient.removeHighLevelProgressListener(this);
	}

	/**
	 * {@inheritDoc}
	 */
	public void nodeConnected(Node node) {
		HighLevelClient highLevelClient = nodeManager.getHighLevelClient(node);
		if (highLevelClient == null) {
			logger.log(Level.WARNING, "got no high-level client for node " + node);
			return;
		}
		try {
			highLevelClient.setWatchGlobal(true);
			getRequests(node);
		} catch (HighLevelException hle1) {
			logger.log(Level.WARNING, "error in backend", hle1);
		} catch (JSiteException jse1) {
			logger.log(Level.WARNING, "error in backend", jse1);
		} catch (IOException e) {
			/* ignore exception, disconnects are handled elsewhere. */
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public void nodeConnectionFailed(Node node, Throwable cause) {
		/* we don't care about this. */
	}

	/**
	 * {@inheritDoc}
	 */
	public void nodeDisconnected(Node node, Throwable throwable) {
		Map<String, Request> identifierRequests = nodeRequests.get(node);
		if (identifierRequests == null) {
			logger.warning("got node without request map: " + node);
			return;
		}
		for (Request request : identifierRequests.values()) {
			fireRequestRemoved(request);
		}
		identifierRequests.clear();
	}

	//
	// INTERFACE HighLevelProgressListener
	//

	/**
	 * @see net.pterodactylus.fcp.highlevel.HighLevelProgressListener#progressReceived(HighLevelClient,
	 *      String, HighLevelProgress)
	 */
	public void progressReceived(HighLevelClient highLevelClient, String identifier, HighLevelProgress highLevelProgress) {
		Node node = nodeManager.getNode(highLevelClient);
		if (node == null) {
			logger.warning("got high-level client without node: " + highLevelClient);
			return;
		}
		Map<String, Request> identifierRequests = nodeRequests.get(node);
		if (identifierRequests == null) {
			logger.warning("got node without request map: " + node);
			identifierRequests = Collections.synchronizedMap(new HashMap<String, Request>());
			nodeRequests.put(node, identifierRequests);
		}
		Request request = identifierRequests.get(identifier);
		if (request == null) {
			logger.warning("got progress for unknown request: " + identifier);
			request = new Request(node, identifier);
			identifierRequests.put(identifier, request);
			fireRequestAdded(request);
		}
		if (highLevelProgress.isFinished()) {
			request.setFinished(true);
			request.setSuccessful(!highLevelProgress.isFailed());
			fireRequestFinished(request);
		} else if (highLevelProgress.isFetchable()) {
			/* TODO - ignore? */
		} else if (highLevelProgress.getURI() != null) {
			request.setURI(highLevelProgress.getURI());
			fireRequestGeneratedURI(request, highLevelProgress.getURI());
		} else {
			request.setTotalBlocks(highLevelProgress.getTotalBlocks());
			request.setRequiredBlocks(highLevelProgress.getRequiredBlocks());
			request.setSuccessfulBlocks(highLevelProgress.getSuccessfulBlocks());
			request.setFailedBlocks(highLevelProgress.getFailedBlocks());
			request.setFatallyFailedBlocks(highLevelProgress.getFatallyFailedBlocks());
			request.setTotalFinalized(highLevelProgress.isTotalFinalized());
			fireRequestProgressed(request);
		}
	}

}
